==> pathena-0.3.3.tgz: .pathena/pytk_client/file_display.py <==

# Support functions are provided to display files by filling a text widget
# with file contents extracted by file type.  Care is taken to identify
# file types and offer the user options to decode unindexed files or
# invoke external viewers.


import os, sys, string, glob
from time import strftime, localtime, asctime

from client_globals import *
from widget_utils   import *

from file_type_attribs import *
from db_operations     import *
from find_state        import many_threshold

max_display_len   = user_prefs.getint('Client', 'max_display_len')

starting_display_position = 5.0
tag_margin = 4

display_fonts = {'sans':  text_font,
                 'serif': serif_text_font,
                 'fixed': small_fixed_font}


#-----------------------------------------

# Note: these procedures were originally methods of QP_GUI_elements.

# The following function is used to insert file text as appropriate in
# the display window (lower panel if attached=true). The file's path
# and type are retrieved from self.path_list using item_num as an index.
# After that, processing proceeds based on file type and other
# attributes.  If text conversion is requested by the user, it will be
# cached in the .pathena/tmp directory in case the file is selected again.

def fill_file_display(self, content_view, attached, item_num,
		      force_display, title_text):
        self.starting_display_position = starting_display_position
        full_path, file_type = self.path_list[item_num][1:3]
        base_path, extension = os.path.splitext(full_path)
        extension = extension[1:]
        type_class = file_type_class[file_type]
        if attached:
            display = self.display
            display.delete('1.0', END)
            display.configure(font=display_fonts[type_class.font_class])
        else:
            display = scrolled_text_display(
                          title='Pathena - %s' % full_path, geometry='+0+0',
                          width=80, height=50, barwidth=11,
                          font=display_fonts[type_class.font_class])
	    display.configure(cursor='', padx=0, pady=0)
        if not os.path.exists(full_path):
            display.insert(END, 'Path:   %s\n\n     '
                           '***** File no longer exists *****\n\n' % full_path,
                           'file_header')
            display.tag_configure('file_header', lmargin1=tag_margin,
                                  background=colors['find_error_bg'])
            raise DeletedFileError

        # Derive file attributes and format a header to display at the
	# top of the text widget.
	
        file_len = os.path.getsize(full_path)
        mod_time = os.path.getmtime(full_path)
        attribs = "Path:   %s\nType:  %s     Last mod:  %s\n" \
                  "Length:  %s (%d)      " % \
                  (full_path, file_type, mod_time_ago(mod_time),
                   approx_size(file_len), file_len)
        display.insert(END, attribs, 'file_header')
        display.mark_set('excerpted_length', display.index('end - 2 char'))
	display.insert(END, "Lines displayed:  \n", 'file_header') 
        display.mark_set('line_count', display.index('end - 2 char'))

        indexed_mod_time = getIndexedModTime(full_path)
        if indexed_mod_time != mod_time:
            out_of_date = '*** File has changed; indexed mod time:  %s\n' % \
                          mod_time_ago(indexed_mod_time)
            display.insert(END, out_of_date, 'file_header')
            self.starting_display_position += 1.0

        display.insert(END, '\n', 'color_separator')
#        display.insert(END, '\n', 'text_separator')
        extend_header_point = display.index('end - 3 char')
        extract_point = display.index('end - 1 char')  ### need this???
        display.tag_configure('file_header', foreground=colors['item_meta_fg'],
                              background=colors['headline_select'],
                              font=text_font)
        display.tag_configure('color_separator', font=spacer_font,
                              background=colors['headline_select'])
#        display.tag_configure('text_separator', font=spacer_font)

        def external_view_buttons(type_class=type_class):
            viewers = type_class.external_viewers
            if not viewers: return
	    display.insert(END, '\n')
            display.insert(END, 'External Viewers:  ', 'metatext')
	    button_width = max([ len(v[0]) for v in viewers ]) - 1
            for name, command in viewers:
                display.insert(END, '  ')
                cmd = lambda c=command: os_system(c % full_path, 2, 1)
                display.window_create(END,
                    window=Button(display, text=name, command=cmd,
                                  width=button_width, pady=2))
            display.insert(END, '\n\n')

        def display_content_insert(content, cached_state=0):
	    display.insert(END, '\n')
            begin = float(display.index('end - 1 char'))
            display.insert(END, content, 'requested_insert')
            delta = int(float(display.index('end linestart')) - begin - 1)
	    if cached_state: line_count = '%s  (cached)' % delta
	    else:            line_count = str(delta)
            display.insert('line_count', line_count)
	    display.tag_configure('requested_insert', rmargin=tag_margin,
				  lmargin1=tag_margin, lmargin2=tag_margin)

        def display_content_delayed():
	    range = display.tag_ranges('content_delayed')
	    display.tag_delete('content_delayed')
	    display.delete(*range)
	    content = decompress_convert(type_class, full_path)[-1]
	    display_content_insert(content)
	    temp_file = new_temp_file_name()
	    write_text_file(content, temp_file)
	    self.displayed_before[item_num] = temp_file
		    
        def display_content(content, cached_state):
	    if cached_state != 2:
	        display_content_insert(content, cached_state)
		return
	    elif self.displayed_before[item_num]:
	        display_content_insert(
		    read_text_file(self.displayed_before[item_num]), 2)
		return
	    start_delayed = display.index(END)
	    display.insert(END, "\nThis file's content has not been indexed "
			   "or cached.\n\n"
			   "  It may be converted and displayed now:   ")
	    display.window_create(END,
		window=Button(display, text='Display Text',
			      command=display_content_delayed,
			      padx=8, pady=2))
	    display.tag_add('content_delayed', start_delayed,
			    display.index(END))
	    display.tag_configure('content_delayed', font=text_font,
				  foreground=colors['item_meta_fg'])

        indexing_attrib = type_class.file_indexing_attrib
        if indexing_attrib == 'other' and extension in compress_ext_name:
            inner_ext = os.path.splitext(base_path)[1][1:]
            if inner_ext in file_ext_name:
                display.insert(END,
                    "This file was compressed with '%s', but when the file "
                    "was indexed,\nthis compression type was not enabled by "
                    "the relevant profile.\n\n"
                    "File extension '%s' is recognized "
                    "for content indexing as type '%s'.\n"
                    % (compress_ext_name[extension], inner_ext,
                       file_ext_name[inner_ext]),
                    'metatext')
		display_content('', 2)
            else:
                display.insert(END,
                    "This file was compressed with '%s'.\n\nFile "
                    "extension '%s' is not recognized for content indexing.\n"
                    % (compress_ext_name[extension], inner_ext),
                    'metatext')

        elif indexing_attrib == 'other' and not force_display \
           and not content_view and not self.displayed_before[item_num]:
            ### following is vestigial: other no longer for known extensions;
            ### only unknown exts go to other -> all map to other class
            file_cmd_attribs = look_up_file_attribs(full_path)
            file_is_text = 'text' in file_cmd_attribs
            apparent_class = file_ext_class.get(extension)
            if apparent_class and extension:
                display.insert(END, 'Apparent file type: %s\n\n' %
                                    file_ext_name[extension],
                               'metatext')
                external_view_buttons(apparent_class)
            else:
                display.insert(END,
                    "No file type is associated with extension '%s'.\n\n"
                    "Attributes according to 'file' command:\n\n    %s\n" %
                    (extension, file_cmd_attribs),
                    'metatext')
                if not file_is_text:
                    display.insert(END, "\nIt appears this file does not "
                        "contain printable text.\n\n"
                        "Display was not attempted; "
                        "right click to force display.\n",
                        'metatext')
            metadata_class = apparent_class or type_class
            display.insert(END, metadata_class.extract_metadata(full_path),
                           'metatext')
            if file_is_text:
                display.insert(END, '----------------------\n', 'metatext')
                display_content(extract_plain_text(full_path), 0)
            ### can get here with content_view???

        else:
	    cached_state, title_text, metadata_text, extract = \
		get_cached_content_plus(type_class, full_path)
            if extract == '*****':
                extract = '***** An error occurred while converting ' \
                          'this file to text. *****'
            if indexing_attrib == 'listing':
                ### generalize fixed font -- function of file type???
                display_content(extract, cached_state)
                display.tag_configure('listing', font=small_fixed_font) ###
            else:   # text-based files
                if indexing_attrib == 'excerpt':
                    display.insert('excerpted_length',
                        'Excerpted length: %d     ' % len(extract))
                elif indexing_attrib == 'encoded' and extract:
                    display.insert('excerpted_length',
                        'Decoded length: %d     ' % len(extract))
                external_view_buttons()
                if metadata_text:
                    metadata_insert = \
                        '----------------------\n%s----------------------\n\n' 
                    display.insert(END, metadata_insert % metadata_text,
                                   'metatext')
                display_content(extract, cached_state)

        display.tag_configure('metatext', font=text_font,
                              foreground=colors['item_meta_fg'])
        title_text = title_text or type_class.extract_title(full_path) or \
                     os.path.basename(full_path)
        if content_view:
            site_count = \
                highlight_search_terms(self, attached, display, item_num,
				       type_class.identifier_specials)
            if attached:
                self.path_label.configure(text='%s -- %s terms'
                                          % (title_text, site_count))
        else:
            if attached: self.path_label.configure(text=title_text)
        display.tag_add('full_widget', '1.0', END)
        display.tag_configure('full_widget', rmargin=tag_margin,
                              lmargin1=tag_margin, lmargin2=tag_margin)


# All search terms are first highlighted in color.  Locations for each
# term are saved in a list (self.sites) to facilitate scrolling later.
# Terms are located using Tk's regular expression search.  Patterns
# are adjusted as needed for source files and others that use special
# characters in identifiers or words.

def highlight_search_terms(self, attached, display, item_num, specials):
        # certain specials need to be escaped in patterns
        esc_specials = specials.replace('-','\-').replace(']','\]')
        # creates tags actual_<term_num>_<site_num> in display
        # possibly creates tags missing_<term_num>
        def add_missing(count, term_num, term):
            if count: return
            display.insert(END, "\n")
            missing_tag = 'missing_%s' % term_num
            display.insert(END, "No instances of term '%s' exist.\n" % term,
                           missing_tag)
            display.tag_configure(missing_tag, foreground=term_color)
        search_terms = self.headlines[item_num][2][:]   # actual terms
        ### includes negative terms ???
        if not search_terms: return 0
        if attached:
            # display term sites: [ [site,...],... ]
            self.sites = []  # following ensures each null list is appendable
            for i in range(len(search_terms)):  self.sites.append([])
            ### might not need sites: can derive from tag_ranges
        site_count = 0
        term_num = 0
        next = self.starting_display_position   # skip file metadata   # '1.0'
        term_color = colors['term_occ_fg']

	def make_new_pattern(term, mode):
	    # mode == 0: pattern anchors a term at word/identifier boundaries
	    # mode == 1: term matches in any context
	    if mode == 1:
                esc_term = regexp_escape(term)
                pattern  = esc_term
            elif term.isalnum():
                esc_term = term
                pattern  = r'(\W|_|^)%s(\W|_|$)' % esc_term
            else:
                esc_term = regexp_escape(term)  ## re.escape(term)
                pattern  = r'([^\w%s]|^)%s([^\w%s]|$)' \
			   % (esc_specials, esc_term, esc_specials)
	    return esc_term, pattern
            
        term = search_terms[0]
        site_num, mode = 0, 0
	esc_term, pattern = make_new_pattern(term, mode)
        while search_terms:
	    # first pattern matches term and context; second only term
	    start = display.search(pattern, next, END, regexp=1, nocase=1)
            if start and site_num < many_threshold:
                start = display.search(esc_term, start, END,
                                       regexp=1, nocase=1)
                tag = 'actual_%s_%s' % (term_num, site_num)
                end = '%s + %s chars' % (start, len(term))
                display.tag_add(tag, start, end)
                display.tag_configure(tag, foreground=term_color)
                next = end
                site_num += 1
                if attached: self.sites[term_num].append(float(start))
	    elif site_num == 0 == mode:
	        mode += 1
	        esc_term, pattern = make_new_pattern(term, mode)
            elif len(search_terms) <= 1:   # terms exhausted
                if attached:
                    add_missing(site_num, term_num, term)
                    site_count += len(self.sites[term_num])
                    return site_count
                else:
                    return None
            else:
                # no more matches with this term or max threshold reached
                if attached:
                    add_missing(site_num, term_num, term)
                    site_count += site_num
                    term_num += 1
                del search_terms[0]
                term = search_terms[0]
                site_num, mode = 0, 0
	        esc_term, pattern = make_new_pattern(term, mode)
                next = self.starting_display_position


def mod_time_ago(mod_time):
    return '%s ago  (%s)' % (approx_duration(time.time() - mod_time),
			     asctime(localtime(mod_time)))

==> pathena-0.3.4.tgz: .pathena/pytk_client/file_display.py <==

# Support functions are provided to display files by filling a text widget
# with file contents extracted by file type.  Care is taken to identify
# file types and offer the user options to decode unindexed files or
# invoke external viewers.


import os, sys, string, glob
from time import strftime, localtime, asctime

from client_globals import *
from widget_utils   import *

from file_type_attribs import *
from db_operations     import *
from find_state        import many_threshold

max_display_len   = user_prefs.getint('Client', 'max_display_len')

starting_display_position = 5.0
tag_margin = 4

display_fonts = {'sans':  text_font,
                 'serif': serif_text_font,
                 'fixed': small_fixed_font}


#-----------------------------------------

# Note: these procedures were originally methods of QP_GUI_elements.

# The following function is used to insert file text as appropriate in
# the display window (lower panel if attached=true). The file's path
# and type are retrieved from self.path_list using item_num as an index.
# After that, processing proceeds based on file type and other
# attributes.  If text conversion is requested by the user, it will be
# cached in the .pathena/tmp directory in case the file is selected again.

def fill_file_display(self, content_view, attached, item_num,
		      force_display, title_text):
        self.starting_display_position = starting_display_position
        full_path, file_type = self.path_list[item_num][1:3]
	is_other = file_type.startswith('other')
	coerced = is_other and self.app.coerced_types.get(full_path)
	if coerced:
	    file_type = coerced
	    displayed_file_type = coerced + ' (coerced)'
	else:
	    displayed_file_type = file_type
        base_path, extension = os.path.splitext(full_path)
        extension = extension[1:].lower()
        type_class = file_type_class[file_type]
        indexed_mod_time = getIndexedModTime(full_path)

        if attached:
            display = self.display
            display.delete('1.0', END)
            display.configure(font=display_fonts[type_class.font_class])
        else:
            display = scrolled_text_display(
                          title='Pathena - %s' % full_path, geometry='+0+0',
                          width=80, height=50, barwidth=11,
                          font=display_fonts[type_class.font_class])
	    display.configure(cursor='', padx=0, pady=0)
        if not os.path.exists(full_path):
            display.insert(END, 'Path:   %s\n'
			   'Type:  %s     Indexed mod time:  %s\n\n     '
                           '***** File no longer exists *****\n\n' % \
			   (full_path, displayed_file_type,
			    mod_time_ago(indexed_mod_time)),
                           'file_header')
            display.tag_configure('file_header', lmargin1=tag_margin,
                                  background=colors['find_error_bg'])
            raise DeletedFileError

        # Derive file attributes and format a header to display at the
	# top of the text widget.
	
        file_len = os.path.getsize(full_path)
        mod_time = os.path.getmtime(full_path)
        attribs = "Path:   %s\nType:  %s     Last mod:  %s\n" \
                  "Length:  %s (%d)      " % \
                  (full_path, displayed_file_type, mod_time_ago(mod_time),
                   approx_size(file_len), file_len)
        display.insert(END, attribs, 'file_header')
        display.mark_set('excerpted_length', display.index('end - 2 char'))
	display.insert(END, "Lines displayed:  \n", 'file_header') 
        display.mark_set('line_count', display.index('end - 2 char'))

        if indexed_mod_time != mod_time:
            out_of_date = '*** File has changed; indexed mod time:  %s\n' % \
                          mod_time_ago(indexed_mod_time)
            display.insert(END, out_of_date, 'file_header')
            self.starting_display_position += 1.0

        display.insert(END, '\n', 'color_separator')
#        display.insert(END, '\n', 'text_separator')
        extend_header_point = display.index('end - 3 char')
        extract_point = display.index('end - 1 char')  ### need this???
        display.tag_configure('file_header', foreground=colors['item_meta_fg'],
                              background=colors['headline_select'],
                              font=text_font)
        display.tag_configure('color_separator', font=spacer_font,
                              background=colors['headline_select'])
#        display.tag_configure('text_separator', font=spacer_font)

        def external_view_buttons(type_class=type_class):
            viewers = type_class.external_viewers
            if not viewers: return
	    display.insert(END, '\n')
            display.insert(END, 'External Viewers:  ', 'metatext')
	    button_width = max([ len(v[0]) for v in viewers ]) - 1
            for name, command in viewers:
                display.insert(END, '  ')
                cmd = lambda c=command: os_system(c % full_path, 2, 1)
                display.window_create(END,
                    window=Button(display, text=name, command=cmd,
                                  width=button_width, pady=2))
            display.insert(END, '\n\n')

	def display_coercion(coerced_type):
	    self.app.coerced_types[full_path] = coerced_type
	    fill_file_display(self, content_view, attached, item_num,
			      force_display, title_text)

	def show_file_types():
	    title = 'Select File Type Coercion'
	    file_types = file_type_class.keys()
	    file_types.sort()
	    state = [None, None]   # toplevel frame, listbox
	    def select_type():
	        display_coercion(state[1].get(state[1].curselection()[0]))
		state[0].destroy()
	    def make_window(fr):
	        state[0] = fr
	        fr2, listbox = scrolled_listbox(fr)
		state[1] = listbox
		for ft in file_types:
		    if ft.endswith('/'): continue   # skip genus only types
		    listbox.insert(END, ft)
		    if ft == 'text/plain':
		        listbox.selection_set(END)
		fr2.pack(padx=10)
		button_row(fr, (('Select Type', select_type, fr),
				('Cancel', lambda : fr.destroy(), 0))
			   )[0].pack(pady=15)
	    pop_up_toplevel_window(make_window, (), title, bottom_margin=0)

        def coerce_type_button():
	    display.insert(END, '\n      ')
	    display.window_create(END,
                    window=Button(display, text='Regard File Type As...',
				  command=show_file_types, pady=2))
            display.insert(END, '\n\n')

        def display_content_insert(content, cached_state=0):
	    display.insert(END, '\n')
            begin = float(display.index('end - 1 char'))
            display.insert(END, content, 'requested_insert')
            delta = int(float(display.index('end linestart')) - begin - 1)
	    if cached_state: line_count = '%s  (cached)' % delta
	    else:            line_count = str(delta)
            display.insert('line_count', line_count)
	    display.tag_configure('requested_insert', rmargin=tag_margin,
				  lmargin1=tag_margin, lmargin2=tag_margin)

        def display_content_delayed():
	    range = display.tag_ranges('content_delayed')
	    display.tag_delete('content_delayed')
	    display.delete(*range)
	    content = decompress_convert(type_class, full_path)[-1]
	    display_content_insert(content)
	    temp_file = new_temp_file_name()
	    write_text_file(content, temp_file)
	    self.displayed_before[item_num] = temp_file
		    
        def display_content(content, cached_state):
	    if cached_state != 2:
	        display_content_insert(content, cached_state)
		return
	    elif self.displayed_before[item_num]:
	        display_content_insert(
		    read_text_file(self.displayed_before[item_num]), 2)
		return
	    start_delayed = display.index(END)
	    display.insert(END, "\nThis file's content has not been indexed "
			   "or cached.\n"
			   "It may be converted and displayed now:\n\n      ")
	    display.window_create(END,
		window=Button(display, text='Display Textual Content',
			      command=display_content_delayed,
			      padx=8, pady=2))
	    display.tag_add('content_delayed', start_delayed,
			    display.index(END))
	    display.tag_configure('content_delayed', font=text_font,
				  foreground=colors['item_meta_fg'])

        indexing_attrib = type_class.file_indexing_attrib
        if indexing_attrib == 'other' and extension in compress_ext_name:
            inner_ext = os.path.splitext(base_path)[1][1:].lower()
            if inner_ext in file_ext_name:
                display.insert(END,
                    "\nThis file was compressed with '%s', but when the file "
                    "was indexed,\nthis compression type was not enabled by "
                    "the relevant profile.\n\n"
                    "File extension '%s' is recognized "
                    "for content indexing as type '%s'.\n"
                    % (compress_ext_name[extension], inner_ext,
                       file_ext_name[inner_ext]),
                    'metatext')
		display_content('', 2)
            else:
                display.insert(END,
                    "This file was compressed with '%s'.\n\nFile "
                    "extension '%s' is not recognized for content indexing.\n"
                    % (compress_ext_name[extension], inner_ext),
                    'metatext')

        elif indexing_attrib == 'other' and not force_display \
           and not content_view and not self.displayed_before[item_num]:
            ### following is vestigial: other no longer for known extensions;
            ### only unknown exts go to other -> all map to other class
            file_cmd_attribs = look_up_file_attribs(full_path)
            file_is_text = file_cmd_attribs.find('text') >= 0
#            file_is_text = 'text' in file_cmd_attribs           ### 2.3
            apparent_class = file_ext_class.get(extension)
            if apparent_class and extension:
                display.insert(END, 'Apparent file type: %s\n\n' %
                                    file_ext_name[extension],
                               'metatext')
                external_view_buttons(apparent_class)
            else:
                display.insert(END,
                    "\nNo file type is associated with extension '%s'.\n\n"
                    "Attributes according to 'file' command:\n\n    %s\n" %
                    (extension, file_cmd_attribs),
                    'metatext')
                if not file_is_text:
                    display.insert(END, "\nIt appears this file does not "
                        "contain printable text.\n\n"
                        "Display was not attempted; "
                        "right click to force display.\n",
                        'metatext')
	    display.insert(END,
                "\nIf this file's content belongs "
		"to one of the known file types, \n"
		"you may coerce its type for display purposes:\n",
		'metatext')
	    coerce_type_button()
            metadata_class = apparent_class or type_class
            display.insert(END, metadata_class.extract_metadata(full_path),
                           'metatext')
            if file_is_text:
                display.insert(END, '----------------------\n', 'metatext')
                display_content(extract_plain_text(full_path), 0)
            ### can get here with content_view???

        else:
	    cached_state, title_text, metadata_text, extract = \
		get_cached_content_plus(type_class, full_path)
            if extract == '*****':
                extract = '***** An error occurred while converting ' \
                          'this file to text. *****'
            if indexing_attrib == 'listing':
                ### generalize fixed font -- function of file type???
                display_content(extract, cached_state)
                display.tag_configure('listing', font=small_fixed_font) ###
            else:   # text-based files
                if indexing_attrib == 'excerpt':
                    display.insert('excerpted_length',
                        'Excerpted length: %d     ' % len(extract))
                elif indexing_attrib == 'encoded' and extract:
                    display.insert('excerpted_length',
                        'Decoded length: %d     ' % len(extract))
                external_view_buttons()
                if metadata_text:
                    metadata_insert = \
                        '----------------------\n%s----------------------\n\n' 
                    display.insert(END, metadata_insert % metadata_text,
                                   'metatext')
                display_content(extract, cached_state)

        display.tag_configure('metatext', font=text_font,
                              foreground=colors['item_meta_fg'])
        title_text = title_text or type_class.extract_title(full_path) or \
                     os.path.basename(full_path)
        if content_view:
            site_count = \
                highlight_search_terms(self, attached, display, item_num,
				       type_class.identifier_specials)
            if attached:
                self.path_label.configure(text='%s -- %s terms'
                                          % (title_text, site_count))
        else:
            if attached: self.path_label.configure(text=title_text)
        display.tag_add('full_widget', '1.0', END)
        display.tag_configure('full_widget', rmargin=tag_margin,
                              lmargin1=tag_margin, lmargin2=tag_margin)


# All search terms are first highlighted in color.  Locations for each
# term are saved in a list (self.sites) to facilitate scrolling later.
# Terms are located using Tk's regular expression search.  Patterns
# are adjusted as needed for source files and others that use special
# characters in identifiers or words.

def highlight_search_terms(self, attached, display, item_num, specials):
        # certain specials need to be escaped in patterns
        esc_specials = specials.replace('-','\-').replace(']','\]')
        # creates tags actual_<term_num>_<site_num> in display
        # possibly creates tags missing_<term_num>
        def add_missing(count, term_num, term):
            if count: return
            display.insert(END, "\n")
            missing_tag = 'missing_%s' % term_num
            display.insert(END, "No instances of term '%s' exist.\n" % term,
                           missing_tag)
            display.tag_configure(missing_tag, foreground=term_color)
        search_terms = self.headlines[item_num][2][:]   # actual terms
        ### includes negative terms ???
        if not search_terms: return 0
        if attached:
            # display term sites: [ [site,...],... ]
            self.sites = []  # following ensures each null list is appendable
            for i in range(len(search_terms)):  self.sites.append([])
            ### might not need sites: can derive from tag_ranges
        site_count = 0
        term_num = 0
        next = self.starting_display_position   # skip file metadata   # '1.0'
        term_color = colors['term_occ_fg']

	def make_new_pattern(term, mode):
	    # mode == 0: pattern anchors a term at word/identifier boundaries
	    # mode == 1: term matches in any context
	    if mode == 1:
                esc_term = regexp_escape(term)
                pattern  = esc_term
            elif term.isalnum():
                esc_term = term
                pattern  = r'(\W|_|^)%s(\W|_|$)' % esc_term
            else:
                esc_term = regexp_escape(term)  ## re.escape(term)
                pattern  = r'([^\w%s]|^)%s([^\w%s]|$)' \
			   % (esc_specials, esc_term, esc_specials)
	    return esc_term, pattern
            
        term = search_terms[0]
        site_num, mode = 0, 0
	esc_term, pattern = make_new_pattern(term, mode)
        while search_terms:
	    # first pattern matches term and context; second only term
	    start = display.search(pattern, next, END, regexp=1, nocase=1)
            if start and site_num < many_threshold:
                start = display.search(esc_term, start, END,
                                       regexp=1, nocase=1)
                tag = 'actual_%s_%s' % (term_num, site_num)
                end = '%s + %s chars' % (start, len(term))
                display.tag_add(tag, start, end)
                display.tag_configure(tag, foreground=term_color)
                next = end
                site_num += 1
                if attached: self.sites[term_num].append(float(start))
	    elif site_num == 0 == mode:
	        mode += 1
	        esc_term, pattern = make_new_pattern(term, mode)
            elif len(search_terms) <= 1:   # terms exhausted
                if attached:
                    add_missing(site_num, term_num, term)
                    site_count += len(self.sites[term_num])
                    return site_count
                else:
                    return None
            else:
                # no more matches with this term or max threshold reached
                if attached:
                    add_missing(site_num, term_num, term)
                    site_count += site_num
                    term_num += 1
                del search_terms[0]
                term = search_terms[0]
                site_num, mode = 0, 0
	        esc_term, pattern = make_new_pattern(term, mode)
                next = self.starting_display_position


def mod_time_ago(mod_time):
    return '%s ago  (%s)' % (approx_duration(time.time() - mod_time),
			     asctime(localtime(mod_time)))
